﻿/* FMvars.jsx ------ UTF-8 ------------------------------------------------------------------------
                Handle variables more elegantly
  Documentation E:\_DDDprojects\FM-vars\InstallFMvars\scripts\FMvars.pdf
  Attention     Code must be UTF-8. With Windows cp no umlauts etc. are allowed in menu items.
  Comment       To be placed in Startup folder of user area
                Normally compiled using the compiler by running !_Make-distribution
                PrepareInstallation is assumed to be on desktop together with project installation-directory
                This installation program resides in \InstallThis (1 level below PrepareInstallation)
                This utility program resides in \InstallThis\scripts (2 levels below PrepareInstallation)
  Tasks         • The utility script FMpppp resides in \InstallThis\cripts and may be ahk and/or jsx
                • Application scripts will be copied into the FM-user-area and optionally the Adobe script folder
  Reference     Traduction française avec l'aide de https://www.deepl.com/translator par Klaus Daube
  Copyleft      2017 Klaus Daube, Zürich, Switzerland, klaus@daube.ch
  History       2017-11-22  project start
                2017-12-22 v1.0 en only version to Stefan Gentz and  Thomas Becker
                2018-02-09 v1.2 on the server
                2018-03-09 v1.3 on server (installation pgm, project structure)
                2021-09-14 v2.0 start with changing localisation method to XML
*/ ; // ===========================================================================================
//@target framemaker

if (typeof KLD_V == "undefined") {
  KLD_V = {};                                     // Global script object for this suite
}

//@include FMvars/FMvars_Globals.jsx              // all declarations, includes FMvars_Texts.jsx
//@include FMvars/FMvars_Ancillary.jsx            // ancillary functions
//@include FMvars/FMvars_panelV.jsx               // incl. button functions etc.
//@include FMvars/FMvars_panelT.jsx               // incl. button functions etc.
//         FMvars/FMvars_en.xml                   //
//         FMvars/FMvars_de.xml                   // see FMvars_Ancillary.jsx, SwitchUIlanguageX
//         FMvars/FMvars_fr.xml                   //

KLD_V.Main = function () { // =====================================================================
/*            Initial function of this script suite.
Called by     Statement at end of this module
Calling       DebugMenu, , SetUpMenus, SetupNotifications, SwitchUIlanguage
Attention     With functions as objects the call to Main must be at end of this module.
History       2021-09-14
*/
  KLD_V.FM_VERSION = app.VersionMajor;
  Console (KLD_V.APP_NAME + " v" + KLD_V.APP_VERSION);
  KLD_V.APP_LANG = KLD_V.SwitchUIlanguageX (KLD_V.APP_NAME); // etablish the UI texts
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
// --- For debugging you need to switch statements (ESTK not working when menus active)
//                              real work       debugging
  KLD_V.SetupNotifications ();  // uncomment    comment
  KLD_V.SetUpMenus ();          // uncomment    comment
//KLD_V.HandleVariables ();     // comment      uncomment
//KLD_V.DebugMenu ();           // comment      uncomment
} ; // --- end Main  ------------------------------------------------------------------------------

KLD_V.SetUpMenus = function () { // ===============================================================
/*            Define the menu items for document and book
Attention     This function MUST NOT be named DefineMenu - because this is a reserved function!
Called by     main
Calling       Command
Comment       Menu "Special" absent since FM-14 → FormatMenu used instead for both doc/book
              For a list of DDD used Esc shortcuts see E:\_DDDprojects\FM-GUI-modifications\ESC-sequences\FM-ESC-sequences.fm
Reference     Klaus Göbel at https://forums.adobe.com/thread/2236201
              (My method with kb-labels after \t in the menu entry defintions does not work in FM-14)
*/
var docFMvarsMenu, menuLocation, nearbyCmd, oMenus = {}, oCmd = {};
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

// --- Menu entries (could be integrated in to command definitions as strings)
  oMenus.MenuMain      = KLD_V.UItxt.SetUpMenus.@menu00.toString() + KLD_V.APP_VERSION;
  oMenus.MenuDocu      = KLD_V.UItxt.SetUpMenus.@menu01.toString();
  oMenus.MenuVars      = KLD_V.UItxt.SetUpMenus.@menu02.toString();
  oMenus.MenuImpExp    = KLD_V.UItxt.SetUpMenus.@menu03.toString();
// --- Command definition for documents
  oCmd.MenuMain        = DefineCommand(0,"FMvars_Main",      oMenus.MenuMain,   "");
  oCmd.MenuDocu        = DefineCommand(1,"FMvars_Docu",      oMenus.MenuDocu,   KLD_V.UItxt.SetUpMenus.@sc01.toString());
  oCmd.MenuVars        = DefineCommand(2,"FMvars_doc",       oMenus.MenuVars,   KLD_V.UItxt.SetUpMenus.@sc02.toString());
  oCmd.MenuImpExp      = DefineCommand(3,"FMvars_docImpExp", oMenus.MenuImpExp, KLD_V.UItxt.SetUpMenus.@sc03.toString());
// --- Shortcut labels
  oCmd.MenuDocu.KeyboardShortcutLabel   = KLD_V.UItxt.SetUpMenus.@scl01.toString();
  oCmd.MenuVars.KeyboardShortcutLabel   = KLD_V.UItxt.SetUpMenus.@scl02.toString();
  oCmd.MenuImpExp.KeyboardShortcutLabel = KLD_V.UItxt.SetUpMenus.@scl03.toString();
// --- Assign commands to menus
  if (KLD_V.FM_VERSION > 13) {
    menuLocation   = app.GetNamedMenu("InsertMenu");
  } else {
    menuLocation   = app.GetNamedMenu("SpecialMenu");
  }
  nearbyCmd = app.GetNamedCommand("Variable");    // define placement, works only for FM ≥ 16
  oCmd.MenuMain.PrevMenuItemInMenu = nearbyCmd;
  docFMvarsMenu  = menuLocation.DefineAndAddMenu("!FMvars_Main", oMenus.MenuMain); // ! required
  docFMvarsMenu.AddCommandToMenu (oCmd.MenuDocu);
  docFMvarsMenu.AddCommandToMenu (oCmd.MenuVars);
  docFMvarsMenu.AddCommandToMenu (oCmd.MenuImpExp);
//
  UpdateMenus();                                  // refresh and actually make the new menu appear
} //--- end SetUpMenus ----------------------------------------------------------------------------

function Command (cmd) { // =======================================================================
/*            Assign functions to menu commands
              This is also called from DebugMenu.
Arguments     cmd: reference number of command from the menu definition
Called by     SetUpMenus, DebugMenu
Calling       OpenHelpFile, HandleVariables,ImportExportVars
Attention     This functions can only have this name (Command)
Comment       Since the panels are per definition 'in front' they can not be influenced to give
              focus back to the document. The user must click into the document to get it in focus
              to be able to enter another shortcut. Of course the focus of the panel is left also
              when reaching for the menu.
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  switch (cmd) {
    case 1:
      KLD_V.OpenHelpFile ();
      break;
    case 2:
      KLD_V.HandleVariables ();
      break;
    case 3:
      KLD_V.ImportExportVars ();
      break;
  }
//Console ("Leaving Command");
} //--- end Command -------------------------------------------------------------------------------

/* Script Registration and unregistration: --------------------------------------------------------
  The script is registered if it contains SetupNotifications and Notify functions
  It stays registered until close of FM.
  Hence during development manual unregistration is necessary before each modification and
  restart of the script.
// ---------------------------------------------------------------------------------------------- */
KLD_V.DebugMenu = function () { //
/*            Invoke individul menu items
Called by     main
Calling       Command,FillDlgList
Comment       Used only during development - no more invoked in productive version
*/
var aMenuItems, listIndex, txtIntro, wDlgDebug;
  aMenuItems = [
  "KLD_V.OpenHelpFile",         // 0
  "KLD_V.HandleVariables",      // 1
  "KLD_V.ImportAndExport"];     // 2
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
  wDlgDebug = new Window("dialog", KLD_V.UItxt.DebugMenu.@msg_01.toString() ,undefined);
  wDlgDebug.preferredSize.width = 450;
  wDlgDebug.location= [200,200];
  wDlgDebug.t1 = wDlgDebug.add ("statictext", undefined, KLD_V.UItxt.DebugMenu.@msg_02.toString(), {multiline:true});
  wDlgDebug.t1.preferredSize.width = 400;
	wDlgDebug.g1 = wDlgDebug.add ("group", undefined, undefined);
  	wDlgDebug.g1.l1 = wDlgDebug.g1.add ("listbox", undefined, undefined);
    wDlgDebug.g1.l1.preferredSize = [350,100];
  KLD_V.FillDlgList (wDlgDebug.g1.l1, aMenuItems);

  wDlgDebug.g1.l1.onDoubleClick = function () {
    listIndex = wDlgDebug.g1.l1.selection.valueOf();
    Command (listIndex + 1);
  };
  wDlgDebug.show();
} //--- end DebugMenu -----------------------------------------------------------------------------

KLD_V.SetupNotifications = function () { // =======================================================
/*            Watch for the following notifications
Called by     main
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  Notification (Constants.FA_Note_PostActiveDocChange, true);
} //--- end SetupNotifications --------------------------------------------------------------------

KLD_V.RemoveMyNotifications = function () { //
/*            Prepare for start of modified script
Called by     -
Reference     Klaus Göbel in https://forums.adobe.com/thread/2174148
              Markus WiedeMaier in https://forums.adobe.com/thread/2276988
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  Notification (Constants.FA_Note_PostActiveDocChange, false);
} //--- end RemoveMyNotifications -----------------------------------------------------------------

function Notify (iNote, oObject, sParam, iParam) { // =============================================
/*            Handle triggered events
Arguments     iNote:  notification id to identify the event.
                      See F_ApiNotification() in FDK Function Ref.
              oObject: current document or book object which generated the event.
              sParam: string, if any, associated with the notification.
              iParam: Integer, if any, associated with the notification. Probably the fcode or bit flags.
Called by     System
Calling       UpdateDialogues, BuildMsg
Attention     - This function must be named Notify
              - Writing to the console in the Notify function is pretty dangerous:
                Opening a panel in the course of Notify may also open the console window
                When then closing the console window (dismissing it) this will also close the panel
                and the panel ID is lost - subsequent reference to it is invalid!
Debugging     Not working in a Notify function: $.bp(…); $.writeln (…);
              These do work:                    Console (…); alert (…); Alert (…);
Comment       - FA_Note_PostActiveDocChange is triggered also by some Pods/dialgues, for example when inserting a table row
                It is also triggered by screen refresh (CTRL+L)
              - Revert to saved triggers 2, 105 and 29 in sequence. Hence 29 is superfluous here.
              - Open a document triggers 105 and 2 which will call UpdateDialogues twice → may create white screen
              - See discussion Trapping "white document window" at https://forums.adobe.com/thread/2276988
*/
  if (KLD_V.oCurrentDoc === undefined) {return;}  // unwanted notification
  if (KLD_V.settings.bIsDebug) {Console ("\nNotify: iNote= " + iNote +  "\n\tsParam= " + sParam + "\n\tiParam= " + iParam + "\n\tobject.Name=  " + oObject.Name + "\n\tobject.ID=  " + oObject.id)};
  switch (iNote) {
    case Constants.FA_Note_PostActiveDocChange:   // 105 active document has changed
      if (oObject.id === KLD_V.oCurrentDoc.id) {
        if (KLD_V.settings.bIsDebug) {Console ("Notify - same object as before, just return");}
        return;
      } else {
        KLD_V.UpdateDialogues (oObject);
      }
      KLD_V.oCurrentDoc = oObject;
      break;

    default:	                                    // unknown notification
      Console (KLD_V.BuildMsg (KLD_V.UItxt.Notify.@msg_01, iNote));// Notify - Pgm error:\n case «%01» not defined.
  }
} //--- end Notify --------------------------------------------------------------------------------

KLD_V.UpdateDialogues = function (oObject) { // ===================================================
/*            Update dialogues for Notify function
Arguments     oObject: the current document (object provided by Notify
Called by     Notify
Calling       ButtonRefreshV
*/
  if (KLD_V.settings.bIsDebug) {
    KLD_V.ZTrace ();
    Console ("    wPalV: " + wPalV + " active? " + wPalV.active);
  }
//KLD_V.oCurrentDoc
  if (typeof  wPalV !== "undefined" && wPalV.active) {
    KLD_V.ButtonRefreshV ();
  }
} //--- end UpdateDialogues -----------------------------------------------------------------------

KLD_V.HandleVariables = function () { // ==========================================================
/*            Find and manipulate system and user variables
Called by     Command
Calling       SaveCurrentLocation, CollectVarsInDoc, GetSystemVars, GetUserVars, RestoreLocation, PaletteVariables
*/
var oLocation;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  KLD_V.oCurrentDoc = app.ActiveDoc;
  if (!KLD_V.oCurrentDoc.ObjectValid()){
    alert (KLD_V.UItxt.General.@DocRequired.toString(), "HandleVariables", true);
    return;
  }
  oLocation = KLD_V.SaveCurrentLocation (KLD_V.oCurrentDoc); // Save cursor location/TR
//Set up arrays for system variables and user variables
//KLD_V.SetDisplay (false);
//$.bp(true);
  KLD_V.CollectVarsInDoc (KLD_V.oCurrentDoc, KLD_V.aoDocVars); // all variables appearing in document
  KLD_V.SetDisplay (true);
  KLD_V.GetSystemVars (KLD_V.oCurrentDoc, KLD_V.asSysVarNames);
  KLD_V.GetUserVars (KLD_V.oCurrentDoc, KLD_V.asUsrVarNames);

  KLD_V.RestoreLocation (KLD_V.oCurrentDoc, oLocation); // Restore  cursor location/TR
  KLD_V.sVarName     = undefined;                 // prevent repeated entry to take last used one
  KLD_V.sVarValue = undefined;
  KLD_V.PaletteVariables ();
} //--- end HandleVariables -----------------------------------------------------------------------

KLD_V.ImportExportVars = function () { // =========================================================
/*            Import or export variables to/from current document
Called by     Command
Calling       DialogueTransfer
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  KLD_V.oCurrentDoc = app.ActiveDoc;
  if (!KLD_V.oCurrentDoc.ObjectValid()){
    alert (KLD_V.UItxt.General.@DocRequired.toString(), "ImportExportVars", true);
    return;
  }
  KLD_V.DialogueTransfer ()
} //--- end ImportExportVars ----------------------------------------------------------------------


// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
//$.bp(true);
KLD_V.Main ();                                          // follow the standard …
