﻿// --- FMvars_Ancillary.jsx ------ UTF-8 ----------------------------------------------------------
// to be included in FMvars.jsx

KLD_V.BuildMsg = function (msgText, parm1) { // ===================================================
/*            Build message text from constant part + variables
Arguments     msgText: may conatin \n (newline), symbolic parameters %01 ... %09
              parm1 ... parm9:  strings replacing the symbolic parameters %01 … %09
Returns       Function value is modified message text
Called by     Notify, GetFindParameters, OpenSourceFile, DialogueTransfer, ButtonTransfer, PaletteVariables, ButtonEditVar, ButtonNewVar, ButtonDelVar
Calling       -
Comment       \r (same as \n) and \t (tab) not working with Default layout behaviour
*/
var j, nArguments, sTextOut = msgText;

  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
  if (typeof msgText == "xml" ) {msgText = msgText.toString();}
// Following two replacements only neccessary for read-in values, however they do not harm direct string parameters
  sTextOut = msgText.replace (/\\n/g , "\n");
  sTextOut = sTextOut.replace (/\\t/g , "  ");    // not working with Default layout behaviour

  nArguments = arguments.length - 1               // handle the symbolic parameters
  if (nArguments > 9) {
    nArguments = 9;                               // %10 would be read as %1
  }
  for (j = 1; j <= nArguments; j++) {
    sTextOut = sTextOut.replace ("%0" + j, arguments[j]);
  }
  return sTextOut;
} //--- end BuildMsg ------------------------------------------------------------------------------

KLD_V.CheckSystemBB = function (sVarName) { // ====================================================
/*            which building blocks are valid?
Called by     PaletteVariables
Calling       IsInMArray
*/
var jVar, aGroups = [], jGroup, nGroups, sGroup, iGroup, aItems = [], jItem, nItems, sItem;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  KLD_V.asSysBlocks.length = 0;                   // clear array for new build-up
  jVar = KLD_V.IsInMArray (KLD_V.aAllowedBB, 0, sVarName); // in which subarray to search
  aGroups = KLD_V.aAllowedBB[jVar][1];            // Group names
  nGroups = aGroups.length;
  for (jGroup = 0; jGroup < nGroups; jGroup++) {
    sGroup = aGroups[jGroup];
    iGroup =  KLD_V.IsInMArray (KLD_V.aBBgroups, 0, sGroup);
    aItems = KLD_V.aBBgroups[iGroup][1];          // Building blocks
    nItems = aItems.length;
    for (jItem = 0; jItem < nItems; jItem++){
      sItem = aItems[jItem];
      KLD_V.asSysBlocks.push(sItem);
    }
  }
} //--- end CheckSystemBB -------------------------------------------------------------------------

KLD_V.CloneObject = function (oTarget, oSource) { // ==============================================
/*            Clone an object
Arguments     oTarget: The object to be created
              oSource: The object to be copied
Called by     SwitchUIlanguage
Comment       This is a polyfill of the JS 6 feature Object.assign
Reference     https://gist.github.com/rafaelrinaldi/43813e707970bd2d77fa
*/
var from, to = oTarget, index = 0, total = arguments.length,
    hasOwnProperty = Object.prototype.hasOwnProperty;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  while(++index < total){   // starting with 1
    from = arguments[index];
    if (from != null) {
      for(var key in from) {
        if (hasOwnProperty.call(from, key)) {
          to[key] = from[key];
        }
      }
    }
  }
  return to;
} //--- end CloneObject ---------------------------------------------------------------------------

KLD_V.DeleteVarInArray = function (sName, aArray) { // ============================================
/*            Delete item in array
Arguments     sName:  name of the item to be removed
              aArray: array in which to operate
Returns       Modified array
Called by     ButtonDelVar
Reference     array.splice(index, n-delete,insert-item1,…,itemX)
*/
var j, nVars;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  nVars = aArray.length;
  for (j = 0; j < nVars; j++) {
    if (aArray[j] == sName) {
      aArray.splice (j, 1);
      break;
    }
  }
} //--- end DeleteVarInArray ----------------------------------------------------------------------

KLD_V.DisplayVariable = function (oDoc, aoDocVars, iVar) { // =====================================
/*            display the indexed variable
              Select variable to visualize the location of it
Arguments     oDoc:      current document
              aoDocVars: array of variable objects
              iVar:      index in array of variable objects
Called by     GoToVariableAny, GoToVarableNamed
Calling       GetPage
Comment       Since variables may occor on any page type display may switch between these → flicker
Reference     GetPage by Rick Quatro, https://forums.adobe.com/thread/2424187
*/
var oPage, oTR;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
//$.bp(true);
  oPage = KLD_V.GetPage (aoDocVars[iVar].VarObj, oDoc);
  if (oPage == undefined) {
    alert (KLD_V.UItxt.DisplayVariable.@msg_01.toString());
    return;
  };
  if (oPage.ObjectValid () === 1) {
    oDoc.CurrentPage = oPage;                     // Go to the page containing the variable.
  }
//  else {
//    oDoc.CurrentPage = oDoc.FirstBodyPageInDoc;   // default in case of error - occurs obviously very often, hence NOK
//  }
  oTR = aoDocVars[iVar].TextRange;
  oDoc.TextSelection = oTR;                       // Variable is selected
  oDoc.ScrollToText(oTR);                         // Selection is viewed
} //--- end DisplayVariable -----------------------------------------------------------------------

KLD_V.DocIsOpen = function (filename) { // ========================================================
/*            Check whether document is already open
Returns       file is open: document object; file not open: null
Called by     ButtonGetVars, ButtonTransfer
Comment       on Windows file names are not case sensitive, hence both forms may be used.
Reference     Rick Quatro
*/
var oDoc, oFile, sFile;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  oFile = File(filename);
  sFile = oFile.fullName.toUpperCase();           // Uppercase the filename for easy comparison.
  oDoc = app.FirstOpenDoc;
  while (oDoc.id) {                               // Loop through the open documents in the session.
  if (File(oDoc.Name).fullName.toUpperCase() === sFile) {// Compare the document’s name with the one we are looking for.
    return oDoc;                                  // The document we are looking for is open.
  }
    oDoc = oDoc.NextOpenDocInSession;
  }
  return null;                                  // Document is not open
} //--- end DocIsOpen -----------------------------------------------------------------------------

KLD_V.FillDlgList = function(oList, asListItems) { // =============================================
/*            Fill dialogue list from global array
Arguments     oList           dialogue list to be filled from the array
              asListItems     array with the list items
Returns       -
Attention     Clearing oList by the standard method for arrays tilts FM in the statement removeAll
              wPalF.g01.lstCatalgF.items.length = 0;       // this is invalid!
Called by     various call-back functions
History       2020-11-10 tested with FMfindRepl
*/
var j, nItems;

  nItems = asListItems.length;
  oList.removeAll ();                             // clear the list
  for (j = 0; j < nItems; j++) {
    oList.add ("item", asListItems[j]);
  }
} //--- end FillDlgList ---------------------------------------------------------------------------

KLD_V.FindIndexOfItem = function (aArray, sSearch) { // ===========================================
/*            Get Index of of particular contents
Arguments     aArray:   array to be searched
              sSearch:  may be part of what to find - only first occurence in array element is found
Returns       Index as function value - or null
Called by     GoToVariable, GoToVariableAny
*/
var j, saLen;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  saLen = aArray.length;
  for (j = 0; j < saLen; j++) {
    if (aArray[j] == sSearch) {
    return j;
    }
  }
  return null;
} //--- end FindIndexOfItem -----------------------------------------------------------------------

KLD_V.GetFindParameters = function (iType, sSearch) { // ==========================================
/*            set up parameters for various find actions
Arguments     iType: what find to be performed
              sSearch: what to search for - used only for certain types
Returns       Parameters for the find method
Called by     GetIndexNearestVariable, CollectVarsInDoc
Calling       BuildMsg
Reference     FDK Function Reference, F_ApiFind(), https://forums.adobe.com/thread/961616
*/
var findParams, propVal;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  findParams = new PropVals() ;
  switch (iType) {
    case 1: // ---------------------------------- find any marker - not used currently
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindObject;
      propVal.propVal.valType = Constants.FT_Integer;
      propVal.propVal.ival = Constants.FV_FindAnyMarker ;
      findParams.push(propVal);
      return findParams;
//    break;
    case 2: // ---------------------------------- find any variable
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindWrap ;
      propVal.propVal.valType = Constants.FT_Integer;
      propVal.propVal.ival = 0 ;  // don't wrap
      findParams.push(propVal);
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindObject;
      propVal.propVal.valType = Constants.FT_Integer;
      propVal.propVal.ival = Constants.FV_FindAnyVariable ;
      findParams.push(propVal);
      return findParams;
//    break;
    case 3: // ---------------------------------- find marker of type sSearch, don't wrap
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindWrap ;
      propVal.propVal.valType = Constants.FT_Integer;
      propVal.propVal.ival = 0 ;  // don't wrap
      findParams.push(propVal);
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindMarkerOfType;
      propVal.propVal.valType = Constants.FT_String;
      propVal.propVal.sval = sSearch;
      findParams.push(propVal);
      return findParams;
//    break;
    case 4: // ---------------------------------- find character format sSearch
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindWrap ;
      propVal.propVal.valType = Constants.FT_Integer;
      propVal.propVal.ival = 0 ;                  // don't wrap
      findParams.push(propVal);
      propVal = new PropVal() ;
      propVal.propIdent.num = Constants.FS_FindCharTag;
      propVal.propVal.valType = Constants.FO_CharFmt;
      propVal.propVal.sval = sSearch;
      findParams.push(propVal);
      return findParams;
//    break;
    default:
      alert (KLD_V.BuildMsg (KLD_V.UItxt.GetFindParameters.@msg_01, iType), "GetFindParameters", true);
      return null;
  }
} //---  end GetFindParameters --------------------------------------------------------------------

KLD_V.GetCharFmtsList = function (oDoc, sArray) { // ==============================================
/*            Fill array with list of character formats
Arguments     oDoc:     current document
              sArray:   array to be filled
Called by     RefreshGlobals
*/
var chrFormat = oDoc.FirstCharFmtInDoc;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  sArray.length = 0;                              // clear the list
  sArray.push ("</>");                            // Default ¶ format
  while(chrFormat.ObjectValid() == true) {
    sArray.push ("<" + chrFormat.Name + ">");
    chrFormat = chrFormat.NextCharFmtInDoc;
  }
} //--- end GetCharFmtsList -----------------------------------------------------------------------

KLD_V.GetIndexNearestVariable = function (aArray) { // ============================================
/*            Get nearest following variable
Arguments     aArray:  global array of doc variables
Returns       Function value =  index in array
Called by     RefreshGlobals
Calling       GetFindParameters
Comment       On entry we are at oDoc.TextSelection and need to find the index in the array.
              Have not found a method to stay at an alredy selected variable
*/
var oTR, oFindParams, j, maLen, aMarkerRange = [], jMax, oTextItem, oTextLoc, oThisMarker, thatId, thisId;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  if (aArray.length == 1) {
  return 0;}              // only 1 variable - so take this
  oTextLoc = KLD_V.oCurrentDoc.TextSelection.beg; // at possible selection
  oFindParams= KLD_V.GetFindParameters (2, null);       // define the variable search
  oTR = KLD_V.oCurrentDoc.Find  (oTextLoc, oFindParams); // finds text range (object pgf), lenght 1
  aVarsRange = KLD_V.oCurrentDoc.GetTextForRange (oTR, Constants.FTI_VarBegin);  // this is an array
  maLen = aVarsRange.length;
  if (maLen === 0) {
    return null;                                  // no selection, assume 0 in calling routine
  }
  for (j = 0; j < maLen; j++) {                   // loop should not be necessary due to findparms
    oTextItem = aVarsRange[j];                    // superfluous ?
    oThisVar = oTextItem.obj;                     // should contain only one variable
  }
  thisId = oThisVar.Unique;
  jMax = aArray.length;
  for (j = 0; j <  jMax; j++) {                   // find the equivalent item in array
    thatId = aArray[j].VarObj.Unique;
    if (thisId == thatId) {break;}
  }
  return j;
} //--- end GetIndexNearestVariable ---------------------------------------------------------------

KLD_V.GetPage = function (obj, oDoc) { // =========================================================
/*            Get the page with the object on it
Arguments     obj:  object (variable, ...) for which to find the page
              oDoc: document with the object
Returns       Function value =  index object oPage
Called by     DisplayVariable, WriteVariables
Reference     GetPage by Rick Quatro, https://forums.adobe.com/thread/2424187
*/
var frame = 0, cell = 0, objType = "", prop = 0;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  while (obj) {
    frame = obj;
    objType = obj.constructor.name;
    switch (objType) {
      case "SubCol" :
        obj = obj.ParentTextFrame;
        break;
      case "Tbl" :
        obj = obj.FirstRowInTbl.FirstCellInRow;
        break;
      case "Row" :
        obj = obj.FirstCellInRow;
        break;
      case "Cell" :
      case "Pgf" :
      case "AFrame" :
        obj = obj.InTextFrame;
        break;
      case "TextLine" :
      case "TextFrame" :
      case "UnanchoredFrame" :
      case "Arc" :
      case "Ellipse" :
      case "Group" :
      case "Inset" :
      case "Line" :
      case "Math" :
      case "Polygon" :
      case "Polyline" :
      case "Rectangle" :
      case "RoundRect" :
        if (obj.FrameParent.ObjectValid()) {
            obj = obj.FrameParent;
        } else {
            obj = 0;
        }
        break;
      case "XRef" :
      case "Var" :
        prop = oDoc.GetTextPropVal (obj.TextRange.beg, Constants.FP_InTextObj); // FP_InTextObj = 1000
/*      if (prop.propIdent.name == "") {             // D+DD assumption for invalid prop
          obj = 0;
          Console ("GetPage - prop.propIdent.name == ''; KLD_V.indexDocVars= " . KLD_V.indexDocVars);  // undefined, but 0
          break;
        } */
        obj = prop.propVal.obj;
        break;
      default:                                    // Prevent endless loop if unknown object type is found.
        obj = 0;
        break;
    }
  }
  if (frame) {
    return frame.PageFramePage;
  } else {
    return 0;
  }
} //--- end GetPage -------------------------------------------------------------------------------

KLD_V.GetScriptPath = function () { // ============================================================
/*            Get the script path
Arguments     -
Returns       Path of the script in the function
Called by     OpenHelpFile, OpenPdfAtDestination, ButtonHelpT, ButtonHelpV
*/
var lastBSlash, sScriptPath, sWinPath;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  sScriptPath = $.fileName;                       // script path (including the file-name of the script)
                              // platform independent format e.g. /e/_DDDprojects/FM+EN-escript/escript
  sWinPath = new File (sScriptPath);
  sScriptPath = sWinPath.fsName;                  // get Windows format
  lastBSlash = sScriptPath.lastIndexOf("\\");     // \ needs escaping
  sScriptPath = sScriptPath.substring(0, lastBSlash);
  return sScriptPath;                             // no final \ !
} //--- end GetScriptPath -------------------------------------------------------------------------

KLD_V.CollectVarsInDoc = function (oDoc, aDocVars) { // ===========================================
/*            Collect variables in the 'natural' order
Arguments     oDoc:        current document
              aDocVars:    array to be filled with var objects
Called by     HandleVariables, ButtonRefreshV, ButtonEditVar, ButtonNewVar, ButtonDelVar, ButtonInsVar
Calling       SaveCurrentLocation, RestoreLocation, GetFindParameters, InitFA_errno, oVariable
Comment       Variables are found on all page types (Body, Master, Reference).
              Starting points must be oDoc.FirstBodyPageInDoc, .FirstMasterPageInDoc, .FirstRefPageInDoc
*/
var aVarsRange = [], bIsSystem, findParams, jFound = 0, jPageType, iVrLen, oFirst, oPage, oPara,
    oSavedLoc, oTextItem,  oTR = new TextRange(), oVar, sVarName, sValue,
    aPages = [oDoc.FirstBodyPageInDoc, oDoc.FirstMasterPageInDoc, oDoc.FirstRefPageInDoc];
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  oSavedLoc = KLD_V.SaveCurrentLocation (oDoc);
  aDocVars.length = 0;                            // clear array for refill
  if (oDoc == null) {                             // emergency bailing out
    if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ("CollectVarsInDoc - Exit 1");}
    return;
  }
  for (jPageType = 0; jPageType < 3; jPageType++) { // for all page types
    oFirst = aPages[jPageType];
    if (oFirst.id == 0 && jPageType != 1) {continue}; // only master page(s) exist
    oDoc.CurrentPage = oFirst;
    oPage = oDoc.CurrentPage;
    oPara = oPage.PageFrame.FirstGraphicInFrame.FirstPgf;
    while (oPara == undefined) {                  // can happen on ref pages
      oPage = oPage.PageNext;
      if (oPage.id == 0) {
        KLD_V.nDocVars = aDocVars.length;
        KLD_V.RestoreLocation (oDoc, oSavedLoc);
//      KLD_V.WriteVariables ("CollectVarsInDoc 1");    // debug
        if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ("CollectVarsInDoc - Exit 2");}
        return;}                                  // no next page available
      oPara = oPage.PageFrame.FirstGraphicInFrame.FirstPgf;
    }
    oTR   = oDoc.TextSelection;
    oTR.beg.obj =  oTR.end.obj = oPara;
    oDoc.TextSelection = oTR;
    oTR.beg.offset = oTR.end.offset = 0;
    findParams = KLD_V.GetFindParameters (2, null);     // finding any variable, no wrap
    KLD_V.InitFA_errno ();                              // reset - it is write protected
    oTR = oDoc.Find(oTR.beg, findParams);         // and do an initial find to get started.

    while(FA_errno === Constants.FE_Success) {
      jFound += 1;
      oDoc.TextSelection = oTR;                   // set up the text range
      aVarsRange = oDoc.GetTextForRange (oTR, Constants.FTI_VarBegin);  // this is an array
      iVrLen = aVarsRange.length;                 // have not found anything other than 1 to be correct
      if (iVrLen === 0) {                         // Hickup ?
        KLD_V.InitFA_errno ();                          // track the progress
        oTR = oDoc.Find(oTR.beg, findParams);     // prepare for next find
        continue;                                 // while
      }
      oTextItem = aVarsRange[0];                  // have not found anything other than 1 to be correct
      oVar = oTextItem.obj;
      sVarName = oVar.VarFmt.Name;                // name of the variable
      sValue   = oVar.VarFmt.Fmt;                 // contents of variable
      oTR      = oVar.TextRange;                  // the text range of the variable
      bIsSystem= (oVar.VarFmt.SystemVar !== Constants.FV_VAR_USER_VARIABLE);
      aDocVars.push (new KLD_V.oVariable (sVarName, sValue, oVar, bIsSystem, oTR));
      KLD_V.InitFA_errno ();                            // track the progress
      oTR = oDoc.Find(oTR.end, findParams);       // prepare for next find
    }
  }
  KLD_V.nDocVars = KLD_V.aoDocVars.length;
  KLD_V.RestoreLocation (oDoc, oSavedLoc);
//KLD_V.WriteVariables ("CollectVarsInDoc 3");          // debug
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ("CollectVarsInDoc - Exit normal");}
} //--- end CollectVarsInDoc ----------------------------------------------------------------------

KLD_V.CopyArray = function (aTarget, aSource) { // ================================================
/*            Create a duplicate of an array
Arguments     aTarget: array to receive the copy
              aSource: source array
Returns       aTarget
Comment       currently not used
Reference     http://www.2ality.com/2012/12/clear-array.html tells:
              target = []; Creates extra garbage: existing array not reused, a new one is created.
              target.length = 0;  JavaScript semantics dictate that all elements must be deleted; costs time
*/
var j, jMax;
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  jMax = aSource.length;
  aTarget.length = 0;                             // clear array
  for (j=0; j < jMax; j++) {
    aTarget.push(aSource[j]);
  }
} //--- end CopyArray -----------------------------------------------------------------------------

KLD_V.GetSystemVars = function (oDoc, asSysVarNames) { // =========================================
/*            Collect system variables from document into global array
Arguments      oDoc:          current document
               asSysVarNames: array receiving the system variable names
Returns        asSysVarNames sorted
Called by     HandleVariables, ButtonRefreshV, ButtonEditVar, RbSystemVars
Comment       Certain variables (Running HF xx, Current page #) are only valid if we are on a master page
              oVar.Fmt [string] defines the variable contents (building blocks and text strings
              used to create a variable instance with the variable format.
Used in       FMvars
*/
var oPage, oVar, pageType, sVarName;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  oPage = oDoc.CurrentPage;
//pageType = oPage.constructor.name               // BodyPage, MasterPage, RefPage [Rick Quatro]
  pageType = oPage.type                           // 5: body; 6: master; 7: reference
  asSysVarNames.length = 0;                       // clear before refill
  oVar = oDoc.FirstVarFmtInDoc;                   // variable in doc-catalogue
  while (oVar.ObjectValid()) {
    if (oVar.SystemVar !== Constants.FV_VAR_USER_VARIABLE) { // system variable ?
      sVarName = oVar.Name;
      if (pageType != Constants.FO_MasterPage) {  // skip specials on body/ref pages
        if (sVarName.indexOf("Running H/F") == 0 || sVarName.indexOf("Current Page") == 0) {
          oVar = oVar.NextVarFmtInDoc;
          continue;
        }
      }
      asSysVarNames.push(sVarName);               // collect in array
    }
    oVar = oVar.NextVarFmtInDoc;
  }
  asSysVarNames.sort ();                          // #, 1, A, a, …
} //--- end GetSystemVars -------------------------------------------------------------------------

KLD_V.GetUserVars = function (oDoc, asUsrVarNames) { // ===========================================
/*            Collect user variables from document into global array
Returns       modfified asUsrVarNames sorted
Called by     HandleVariables, ButtonRefreshV, ButtonEditVar, RbUserVars
Comment       oVar.Fmt [string] defines the variable contents (building blocks and text strings
              used to create a variable instance with the variable format.
Used in       FMvars
*/
var oVar, sVarName;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  asUsrVarNames.length = 0;                 // clear before refill
  oVar = oDoc.FirstVarFmtInDoc;                   // variable in doc-catalogue
  while (oVar.ObjectValid()) {
    if (oVar.SystemVar === Constants.FV_VAR_USER_VARIABLE) { // user variable ?
      sVarName = oVar.Name;
      asUsrVarNames.push(sVarName);         // collect in array
    }
    oVar = oVar.NextVarFmtInDoc;
  }
  asUsrVarNames.sort ();                    // #, 1, A, a, …
} //--- end GetUserVars ---------------------------------------------------------------------------

KLD_V.VarIsInCatalogue = function (oDoc, sVarName) { // ===========================================
/*            Check whether variable is in catalogue
Returns       VarFmt object if var is already in catalgoue, else null
Called by     GoToVariable, GoToVariableAny, ButtonDelVar
*/
var oVar;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  oVar = oDoc.FirstVarFmtInDoc;                   // variable in doc-catalogue
  while (oVar.ObjectValid()) {
    if (sVarName == oVar.Name) {
      return oVar;                                // already in catalogue
    }
    oVar = oVar.NextVarFmtInDoc;
  }
  return null;
} //--- end VarIsInCatalogue ----------------------------------------------------------------------

KLD_V.GetVarNames = function (oDoc, aNames, aoSourceVars) { // ====================================
/*            Collect variables from the catalogue
Arguments     oDoc:         current document
              aNames:       array to be filled with var names: user vars alphabetically, then system vars alphabetically
              aoSourceVars: array of variable objects (name, value)
Called by     ButtonGetVars
Comment       Variables are found in the catalogue in an unpredictable order
              Since both name and content of the variables are needed for transfer,
              array aoSourceVars is set up in parallel
*/
var j, oPage, oVar, pageType, sVarName, aoUsrVars = [], nUsrNames, aoSysVars = [], nSysNames;
function fSort (a, b) {
  if      (a.Name >  b.Name) {return 1;}
  else if (a.Name == b.Name) {return 0;}
  else                       {return -1;}
} //--- end fSort --------------------------------
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  aNames.length = 0;                              // clear before refill
  aoSourceVars.length = 0;
  oVar = oDoc.FirstVarFmtInDoc;                   // variable in doc-catalogue
  while (oVar.ObjectValid()) {
    sVarName = oVar.Name;
    sVarValue = oVar.Fmt;
    if (oVar.SystemVar === Constants.FV_VAR_USER_VARIABLE) { // user variable ?
      aoUsrVars.push (new KLD_V.oVariable(sVarName, sVarValue));
    } else {                                      // system variable
      aoSysVars.push (new KLD_V.oVariable(sVarName, sVarValue));
    }
    oVar = oVar.NextVarFmtInDoc;
  }
  aoUsrVars.sort (fSort);                         //  otherwise completely usesess order !
  aoSysVars.sort (fSort);                         // #, 1, A, a, …
  nUsrNames = aoUsrVars.length;
  for (j = 0; j < nUsrNames; j++){                // add user vars
    aoSourceVars.push(aoUsrVars[j]);
    aNames.push(aoUsrVars[j].Name);
  }
  nSysNames = aoSysVars.length;
  for (j = 0; j < nSysNames; j++){                // add system vars
    aoSourceVars.push(aoSysVars[j]);
    aNames.push(aoSysVars[j].Name);
  }
} //--- end GetVarNames ---------------------------------------------------------------------------

KLD_V.GoToVariableAny = function (iVar, bDirection) { // ==========================================
/*            Got to any variable in sequence of object array
Arguments     iVar:       index of variable in array of var objects where we are now
                          null: go to the extreme (first/last)
              bDirection: true: searching forwards (to last), false: searching backwards (to first)
Called by     ButtonFirst, ButtonPrevious, ButtonNext, ButtonLast
Calling       IsVarUsed, DisplayVariable, SyncEditArea, FindIndexOfItem
*/
var bAnyVar, bNotfound = true, bSystemVar, jVar = iVar, jMax, oDoc, sName, sSearch;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  if (iVar == null) {                             // the very first or very last is requested
    jVar = bDirection ? KLD_V.nDocVars-1: 0;
  }
  KLD_V.GoToVariableThis (jVar);                        // 0 … nDocVars - 1
} //---  end GoToVariableAny ----------------------------------------------------------------------

KLD_V.GoToVariableNamed = function (iVar, bDirection) { // ========================================
/*            Got to any variable in sequence of object array
Arguments     iVar:       index of variable in array of var objects where we are now
                          null: go to the extreme (first/last)
              bDirection: true: searching forwards (to last), false: searching backwards (to first)
Called by     ButtonFirst, ButtonPrevious, ButtonNext, ButtonLast
Calling       IsVarUsed, DisplayVariable, SyncEditArea, FindIndexOfItem
Comment       Since variables may occur on any kind of pages, the display may change to any type of page.
*/
var bAnyVar, bNotfound = true, bSystemVar, jVar, jMax, oDoc, sName, sSearch;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

    if (wPalV.g1.pVars.listVarNames.selection == null) {return;} // nothing is selected
    sSearch = wPalV.g1.pVars.listVarNames.selection.text; // selection in the list
    if (!KLD_V.IsVarUsed (oDoc, KLD_V.aoDocVars, sSearch)) {
      wPalV.gCtrl.message.text = KLD_V.UItxt.GoToVariable.@msg_01.toString();
      return;
    }

  jMax  = KLD_V.nDocVars - 1;
  if (iVar == null) {                             // The very first or very last is requested
    if (bDirection) {                             // Find last named
      for (jVar = jMax; jVar >= 0; jVar--) {
        sName = KLD_V.aoDocVars[jVar].Name;
        if (sName == sSearch) {                   // search var found
          bNotfound = false;
          KLD_V.indexDocVars = jVar
          break;
        }
      }
    } else {                                      // Find first named
      for (jVar = 0; jVar <= jMax; jVar++) {
        sName = KLD_V.aoDocVars[jVar].Name;
        if (sName == sSearch) {                   // search var found
          bNotfound = false;
          KLD_V.indexDocVars = jVar
          break;
        }
      }
    }

  } else {                                        // Named variable 0 … jMax is requested
    if (bDirection) {                             // next named to be found
      for (jVar = iVar; jVar <= jMax; jVar++) {
        sName = KLD_V.aoDocVars[jVar].Name;
        if (sName == sSearch) {                   // search var found
          bNotfound = false;
          KLD_V.indexDocVars = jVar
          break;
        }
      }
    } else {
      for (jVar = iVar; jVar >= 0; jVar--) {
        sName = KLD_V.aoDocVars[jVar].Name;
        if (sName == sSearch) {                   // search var found
          bNotfound = false;
          KLD_V.indexDocVars = jVar
          break;
        }
      }
    }
  }
  if (bNotfound ) {
    wPalV.gCtrl.message.text = KLD_V.UItxt.GoToVariable.@msg_02.toString();
    return;
    }
  KLD_V.GoToVariableThis (jVar);
} //---  end GoToVariableNamed --------------------------------------------------------------------

KLD_V.GoToVariableThis = function (iVar) { // =====================================================
/*            Go to the desired variable
Arguments     iVar:       index of variable in array of var objects where to go
Called by     GoToVariableAny, GoToVariableNamed
Calling       IsVarUsed, DisplayVariable, SyncEditArea, FindIndexOfItem
*/
  KLD_V.indexDocVars = iVar;                      // keep for other nav buttons
  oDoc  = KLD_V.oCurrentDoc;
  jMax  = KLD_V.nDocVars - 1;
  sName = KLD_V.aoDocVars[iVar].Name;             // item in doc variable array

  KLD_V.DisplayVariable (KLD_V.oCurrentDoc, KLD_V.aoDocVars, iVar);    // can be on any page type
  wPalV.g1.pCurVar.g3.sVarName.text = sName;      // indicate in GUI
  wPalV.g1.pCurVar.g4.sVarValue.text = KLD_V.aoDocVars[iVar].Value;
  bSystemVar = KLD_V.aoDocVars[iVar].IsSystem;
  KLD_V.SyncEditArea (bSystemVar);                      // fill edit area according to variable found in doc
// Now we have either system vars or user vars in the display list.
  if (bSystemVar) {
    jVar = KLD_V.FindIndexOfItem (KLD_V.asSysVarNames, sName);     // index of item in display list
    if (jVar == null) {return;}
    wPalV.g1.pVars.listVarNames.items[jVar].selected = true; // Indicate var in list
  } else {
    jVar = KLD_V.FindIndexOfItem (KLD_V.asUsrVarNames, sName);     // index of item in display list
    if (jVar == null) {return;}
    wPalV.g1.pVars.listVarNames.items[jVar].selected = true; // Indicate var in list
  }
  KLD_V.indexDocVars = iVar                       // for the other buttons
} //---  end GoToVariableThis ---------------------------------------------------------------------

KLD_V.InitFA_errno = function () { // =============================================================
/*            Reset FA_errno as it is write protected
Called by     CollectVarsInDoc
Reference     https://forums.adobe.com/thread/962910
*/
  app.GetNamedMenu("!MakerMainMenu");             //If this fails, you've got bigger problems
  return;
} //---  end InitFA_errno -------------------------------------------------------------------------

KLD_V.IsInArray = function (array, item) { // =====================================================
/*            Get index of item in array
Arguments     array: where to find the item
              item:  to be found in the array
Returns       Index of found element or null.
Called by     ButtonNewVar
Comment       array.indexOf() is not supported in ExtendScript, only  string.indexOf () is defined.
*/
var j, jMax;
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  jMax = array.length;
  for (j = 0; j < jMax; j++) {
    if (array [j] == item ) { return j;}
  }
  return null;                                    // not found in array
} //--- end IsInArray -----------------------------------------------------------------------------

KLD_V.IsInMArray = function (aArray, mIndex, sItem) { // ==========================================
/*            Get index of item in multidimensional array
Arguments     aArray: where to find the item
              mIndex: In which subarray to search
              sItem:  to be found in the array
Returns       Index of found element or null.
Called by     CheckSystemBB
Comment       array.indexOf() is NOT supported in ExtendScript, only  string.indexOf () is defined.
*/
var j, jMax, sFound;
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  jMax = aArray.length;
  for (j = 0; j < jMax; j++) {
    sFound = aArray [j][mIndex];
    if (aArray [j][mIndex] == sItem ) { return j;}
  }
  return null;                                    // not found in array
} //--- end IsInMArray ----------------------------------------------------------------------------

KLD_V.OpenHelpFile = function () { // =============================================================
/*            Menu item "Documentation"
              Display the pdf with same name as this script ($.filename includes extension)
Called by     Command
Calling       GetScriptPath
Comment       No feedback in case the pdf does not exist → probably to be changed
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  KLD_V.sHelpFile = File (KLD_V.GetScriptPath () + "\\" + KLD_V.APP_NAME + ".pdf"); // abs URI notation
  if (KLD_V.sHelpFile.exists) {                   // KLD_V.sHelpFile
    KLD_V.sHelpFile.execute();
  }
} //--- end OpenHelpfFile -------------------------------------------------------------------------

KLD_V.OpenSourceFile = function (sFilename) { // ==================================================
/*            open file whithout displaying it
Arguments     sFilename:  name(path) of file to open
Returns       Function is doc object
Called by     ButtonGetVars, ButtonTransfer
Calling       BuildMsg
Reference     Book archive version 2 (http://blogs.adobe.com/techcomm/2011/07/create-a-book-packager-using-extendscript.html)
              Anchal, Adobe FrameMaker Engineering Team
*/
var oDoc, i, openProp, openReturnParams;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

	openProp = GetOpenDefaultParams();
	i=GetPropIndex(openProp,Constants.FS_FileIsOldVersion);
	openProp[i].propVal.ival=Constants.FV_DoOK;

	i=GetPropIndex(openProp,Constants.FS_MakeVisible);
	openProp[i].propVal.ival=false;

	i=GetPropIndex(openProp,Constants.FS_FontNotFoundInDoc);
	openProp[i].propVal.ival=Constants.FV_DoOK;

	i=GetPropIndex(openProp,Constants.FS_FileIsInUse);
	openProp[i].propVal.ival=Constants.FV_ResetLockAndContinue;

	i=GetPropIndex(openProp,Constants.FS_AlertUserAboutFailure);
	openProp[i].propVal.ival=Constants.FV_DoCancel;

	i=GetPropIndex(openProp,Constants.FS_LockCantBeReset);
	openProp[i].propVal.ival=Constants.FV_DoOK;

	i=GetPropIndex(openProp,Constants.FS_RefFileNotFound);
	openProp[i].propVal.ival=Constants.FV_AllowAllRefFilesUnFindable;

	openReturnParams = new PropVals();
	oDoc = Open(sFilename, openProp, openReturnParams);
  if (oDoc.ObjectValid() === 1) {
    return oDoc;                                  // Return the document object.
  } else {                                        // Can't open, print the errors to the Console
    Alert (KLD_V.BuildMsg (KLD_C.UItxt.OpenSourceFile.@msg_01, openReturnParams, "OpenSourceFile"), true); //error icon
    return false;
  }
} //--- end OpenSourceFile ------------------------------------------------------------------------

KLD_V.OpenPdfAtDestination = function (pdfFile, location) { // ====================================
/*            Open PDF at specified destination
Arguments     pdfFile:  full path in absoluteURI notation
              location: full definition - not just what was set in FM, e.g. "nameddest=M8.newlink.FMcalcseries"
Called by     ButtonHelpT, ButtonHelpV
Calling       GetScriptPath
Comment       requires OpenPdf.exe in script directory
              (source: E:\_DDDprojects\FM-calc\test-scripts\OpenPdfWithParameters.ahk)
Reference     http://www.adobe.com/content/dam/acom/en/devnet/acrobat/pdfs/pdf_open_parameters.pdf
*/
var path, tempFile;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  path = KLD_V.GetScriptPath () + "\\";
  tempFile = File (path + "FMxxx_tScript.bat");  // Create the calling batch file
  tempFile.open("w");
  tempFile.encoding = "UTF-8";
  tempFile.writeln ("REM Generated by FMvars_Ancillary.jsx, function OpenPdfAtDestination");
  tempFile.writeln (path + "FMxxx_OpenPdf.exe " + '/A ' + location + ' "' + pdfFile.fsName + '"');
  tempFile.close();
  tempFile.execute ();
  tempFile.delete;
} //---  end OpenPdfAtDestination -----------------------------------------------------------------

KLD_V.RemoveTextRange = function (oDoc, oTR) { // =================================================
/*            Remove text range from document text
Arguments     oDoc: Current document
              oTR:  TextRange, current selection
Called by     ButtonRemove
Reference     FindAndDeleteVariable in FMcalc
*/
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  oDoc.TextSelection = oTR;
  oDoc.DeleteText(oTR);                           // Delete the selected text
} //--- end RemoveTextRange -----------------------------------------------------------------------

KLD_V.RestoreLocation = function (oDoc, oLocation) { // ===========================================
/*            Restore currunt location to saved one
Arguments     oDoc:      current document
              oLocation: object created by SaveCurrentLocation
Called by     HandleVariables, CollectVarsInDoc, RefreshGlobals
Attention     oLocation must not be a global variable - or a stack must be established
              Not tested what happens if this document is no more open at restore
Reference     Rick Quatro at https://forums.adobe.com/thread/2243878
*/
var oPgf, tLoc, textRange;
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  oDoc.CurrentPage = oLocation.page;
  oPgf = oLocation.pgf;
  if (oPgf.ObjectValid () === 1) {
    oDoc.TextSelection =   oLocation.txtRange;
  }
//$.bp(true);
  textRange = oLocation.txtRange;
  oDoc.ScrollToText(textRange);
} //--- end RestoreLocation -----------------------------------------------------------------------

KLD_V.SaveCurrentLocation = function (oDoc) { // ==================================================
/*            Save the currunt location for later restore
Arguments     oDoc: current document
Called by     HandleVariables, CollectVarsInDoc, RefreshGlobals
Attention     Not tested what happens if this document is no more open at restore
Reference     Rick Quatro at https://forums.adobe.com/thread/2243878
*/
var oLocation = {};
//$.bp(true);
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  oLocation.pgf    = oDoc.TextSelection.beg.obj;
  oLocation.txtRange = oDoc.TextSelection;
  oLocation.page   = oDoc.CurrentPage;
  return oLocation;
} //--- end SaveCurrentLocation -------------------------------------------------------------------

KLD_V.SetDisplay = function (bWanted) { // ========================================================
/*            Set display to on or off
Arguments     bWanted: true → set display ON; false → display OFF
Called by     CollectVarsInDoc (only potentially)
Commment      see https://forums.adobe.com/thread/2443666 why this is not used
*/
var iWanted = 0, iCurrent = app.Displaying, sStack;

  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
  if (bWanted) {iWanted = 1;}                     // to be able to use strict equal operator
  if (iCurrent !== iWanted) {                     // if current setting not the wanted one
    app.Displaying = iWanted;                     // set to the wanted one
  }
} //--- end SetDisplay ----------------------------------------------------------------------------

KLD_V.StrTrim = function (string) { // ============================================================
/*            Trim spaces etc from beginning or end of string
              Trim spaces, tabs, new lines of various sorts from beginning or end of string
Arguments     string: original string, can not be modified
Returns       Modified string
Called by     ButtonEditVar, ButtonNewVar
Comment       string.trim not supported in ES
Reference     agopaul at https://forums.adobe.com/thread/1974556
*/
var re_Trim = /(^[\s\n\r\t\x0B]+)|([\s\n\r\t\x0B]+$)/g;
  if (KLD_V.settings.bVerbose) {KLD_V.ZTrace ();}

  return string.replace(re_Trim, "");
} //--- end StrTrim -------------------------------------------------------------------------------

KLD_V.IsVarUsed = function (oDoc, aoDocVars, sFindName) { // ======================================
/*            Is the variable used in the document?
Arguments     oDoc:      current document
              aoDocVars: array of variable objects (Name, Value, etc.)
              sVarName:  name of variable to be checked
Returns       true if var is used in doc, else false
Called by     GoToVariable, GoToVariableAny, ButtonDelVar
*/
var jVar, jMax, sVarName;
  if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}

  jMax = aoDocVars.length -1
  for (jVar = 0; jVar <= jMax; jVar++) {
    sVarName = aoDocVars[jVar].Name;
    if (sVarName == sFindName) {
      return true;
    }
  }
  return false;                                   // not found in the used variables
} //--- end IsVarUsed -----------------------------------------------------------------------------

KLD_V.SwitchUIlanguageX = function (sLangFile) { // ===============================================
/*            Read the language relevant XML file to fill the global array of texts
Arguments     sLangFile       Base of the languaage xml file. will be appended
                              _de.xml etc.
                              For test pupose this may be the beginning of a real path name
Returns       Values in global XML object KLD_V.UItxt.
Called by     Main
Calling       GetXMLdata
Comment       I do not convert everything to strings, but use toString() where required
Used in       FMfindRepl
History       2020-11-16 test with VSCode and ESDebugger
*/
var appLang, sLangCode;

//sLangFile = KLD_V.APP_NAME + "\\" + sLangFile;  // need to add subDir here <<<< ??? >>>>
  appLang = app.Language;
//appLang = Constants.FV_LANG_GERMAN;             // set explictly for UI test
	switch (appLang) {
		case Constants.FV_LANG_ENGLISH:               // 0x01
		case Constants.FV_LANG_BRITISH: {             // 0x02
			KLD_V.UItxt = KLD_V.GetXMLdata (sLangFile + "_en.xml");
			sLangCode = "en";
			break;
		}
		case Constants.FV_LANG_GERMAN: {              // 0x03
			KLD_V.UItxt = KLD_V.GetXMLdata (sLangFile + "_de.xml");
			sLangCode = "de";
			break;
		}
		case Constants.FV_LANG_FRENCH: {              // 0x05
			KLD_V.UItxt = KLD_V.GetXMLdata (sLangFile + "_fr.xml");
			sLangCode = "fr";
			break;
		}
		default: {
			KLD_V.UItxt = KLD_V.GetXMLdata (sLangFile + "_en.xml");
			KLD_V.TEMPLATE_FILE = goFMc.TEMPLATE_BASE + "en.fm";
			sLangCode = "en";
		}
	}
	return sLangCode;
} //--- end SwitchUIlanguageX ---------------------------------------------------------------------

KLD_V.GetXMLdata = function (sXmlFile) { // =======================================================
/*            Read localisation strings from xml file
Argements     sXmlFile        fully qualified path of xml file or just the file name.
              File must exist
              If no filepath is given (no slash or backslash present),
              then the file is assumed in the same directory as the script
Returns       xml sructure
Usage         xSettings = getXMLdata (sXmlFile);
Called by     SwitchUIlanguageX
Attention     this function is called by SwitchUIlanguageX, hence localisation has not finished yet.
              Therefore: error messages need to be localised with the ExtendScript method
Reference     E:\FM-ExtendScript\SaveBookOldVersion\SaveBookAsOldVersion.jsx by Rick Quatro
History       2020-11-11
*/
var e, fXmlFile, xData, msg_01, msg_02;

  msg_01 = { en: "XML file «" + sXmlFile + "» not found.",
             de: "XML datei «" + sXmlFile + "» nicht gefunden.",
             fr: "Fichier XML «" + sXmlFile + "» non trouvé."};
  msg_02 = { en: "Read error on XML file «" + sXmlFile + "». Error:\n "+ e,
             de: "Lesefehler auf XML datei «" + sXmlFile + "» . Fehler:\n "+ e,
             fr: "Erreur de lecture sur le fichier XML «" + sXmlFile + "». Erreur:\n " + e};
  if ((sXmlFile == null) || (sXmlFile == undefined)) {
    return false;
  } else if (sXmlFile.indexOf("/") > 0 || sXmlFile.indexOf("\\") > 0) { // fully qualified path
    fXmlFile = File(sXmlFile);
  } else {
    fXmlFile = File($.fileName.replace (/[^\\\/]+$/i , sXmlFile));  // file in script folder
  }
  if (fXmlFile.exists === false) {
      Alert (localize (msg_01), "getXMLdata", true);
      return false;
  }
  fXmlFile.open("r");
  fXmlFile.encoding = "UTF-8";
  try {
      xData = new XML(fXmlFile.read());
      fXmlFile.close();
      return  xData;
  } catch (e) {                                   // what kind of error can this be ?
      Alert (localize (msg_02), "getXMLdata", true);
      fXmlFile.close();
      return false;
  }
} //--- end getXMLdata ----------------------------------------------------------------------------


KLD_V.WriteVariables = function (sComment) { // ===================================================
/*            write out found variables
Calling       GetPage
*/
var j, oPage, pageType, pageName, oDoc = KLD_V.oCurrentDoc;

  $.writeln ("\n" + oDoc.Name + " Comment = " + sComment);
  for (j = 0; j < KLD_V.nDocVars; j++) {
    oPage = KLD_V.GetPage (KLD_V.aoDocVars[j].VarObj, KLD_V.oCurrentDoc)
    pageType = oPage.type;                          // 5: body; 6: master; 7: ref
    pageName = oPage.Name;
    if (pageType == 5) {pageName = oPage.PageNum};
    $.writeln (j + "\t" + KLD_V.aoDocVars[j].VarObj.Unique+ "\t" + KLD_V.aoDocVars[j].IsSystem + "\t" + pageName+ "\t" + KLD_V.aoDocVars[j].Name);
  }
//$.bp(true);
} //--- end WriteVariables ------------------------------------------------------------------------

KLD_V.ZTrace = function (msg) { // ================================================================
/*            Indicate calling sequence
Argumens      msg: if present, only this is issued with appropriate indentation
Called by     all other functions for debugging purpose: if (KLD_V.settings.bIsDebug) if (KLD_V.settings.bIsDebug) {KLD_V.ZTrace ();}
*/
var indent, name, level, sStack, aStack = []
  , dots =   ". . . . : . . . . : . . . . : . . . . : . . . . : . . . . : . . . . : . . . . : . . . . :"
  , blanks = "                                                                                         ";

  sStack = $.stack.toString();
  aStack = sStack.split ("\n");
  level = aStack.length - 3;
  if (level >= 0) {
    if (msg == undefined) {
      name = aStack[level];
      indent = dots.substring(0, level*2);
      Console (indent + name);
    } else {
      indent = blanks.substring(0, level*2);
      Console (indent + msg);
    }
  } else {
    Console ("ZTrace: level < 0 - Stack:\n" + $.stack);
  }
} // --- end ZTrace
