﻿/* --- FMvars_Globals.jsx ------ UTF-8 ------------------------------------------------------------

   Calling   There must not be any called functions here.
   Comment   to be included in FMvars.jsx
   Reference "JavaScript: The Good Parts" by Douglas Crockford [O’Reilly, 2008-05]
             is not a friend of global variables, hence I switched to object definition
   History   2021-09-14
*/ ; // ===========================================================================================

  KLD_V.UItxt = {};                               // current texts by language switching

// === Windows ID must be set outside the dialogue function to be available for Notify functions
//     wPalV  = new Window("palette", undefined, undefined);  // title set in palette
//     var declarations for the global windows variables left out to make it Superglobal
//     See E:\_DDDprojects\_JSX-library\_EscriptSpecials.fm, Superglobal

// === Constants (per definition global), global variables, arrays ================================
// const works only in ESTK. Invoking the script in FM I get "XXX_YYY_ZZ redeclared" for all these items
  KLD_V.APP_NAME     = "FMvars";                  // the name of the game
  KLD_V.APP_VERSION  = "2.1";                     // current version of script (inc after publication)
  KLD_V.APP_LANG     = undefined;                 // set in SwitchUIlanguageX
  KLD_V.FM_VERSION  = 15;                         // set in Main

  KLD_V.oCurrentDoc  = undefined;                 // Currently active document
  KLD_V.oCurrentVar  = undefined;                 // Current variable as selected in text
  KLD_V.aoDocVars    = [];                        // Array of doc Variable objects (see oVariable)
  KLD_V.indexDocVars = undefined;                 // Index of navigated doc variable in aDocVars
  KLD_V.nDocVars     = undefined;                 // Number of doc variables
  KLD_V.bIsUserVar   = undefined;                 // false for system vars
  KLD_V.bGoNamed     = undefined;                 // false for walking through any variable
  KLD_V.asSysVarNames= [];                        // array of system variable names
  KLD_V.nSysVars     = undefined;                 // number of system variables in document
  KLD_V.asUsrVarNames= [];                        // array of user variable names
  KLD_V.nUsrVars     = undefined;                 // Number of user variables in document
  KLD_V.sVarName     = undefined;                 // Name of current variable in the variable list
  KLD_V.sVarValue    = undefined;                 // Contents of current variable in the variable list
  KLD_V.asCharFmts   = [];                        // Character formats in document
  KLD_V.sHelpFile    = undefined;                 // defined in OpenHelpFile
  KLD_V.bImport      = undefined;                 // true for import, false for export of variables
  KLD_V.sFileName    = undefined;                 // 'external' file
  KLD_V.aoSourceVars = [];                        // array of variable objects from catalogue
  KLD_V.aoTransVars  = [];                        // array of variable objects (name, value)
  KLD_V.oSourceDoc   = undefined;                 // transfer source
  KLD_V.oTargetDoc   = undefined;                 // transfer target

// === Building blocks (BBs)for system variables in the order as provided by FM ===================
  KLD_V.asSysBlocks = [];                         // Collection of allowed BBs
  KLD_V.asSysBlocksAll  =                         // Available BBs - order is not important for the script
                                                  // If these items change, also the other arrarys must be modified
  ["<$marker1>", "<$marker2>", "<$marker3>", "<$marker4>", "<$marker5>", "<$marker6>", "<$marker7>", "<$marker8>",
   "<$paratext[paratag]>", "<$paranum[paratag]>", "<$paranumonly[paratag]>", "<$paratag[paratag]>", "<$condtag [hitag,...,lotag,nomatch]>",
   "<$second>", "<$second00>", "<$minute>", "<$minute00>", "<$hour>", "<$hour01>", "<$hour24>", "<$ampm>", "<$AMPM>", "<$daynum>", "<$daynum01>",
   "<$dayname>", "<$shortdayname>", "<$monthnum>", "<$monthnum01>", "<$monthname>", "<$shortmonthname>", "<$year>", "<$shortyear>",
   "<$currenttime>", "<$modificationtime>", "<$creationtime>",
   "<$daynumkanjikazu>", "<$daynumkanjinumeric>", "<$monthnumkanjikazu>", "<$monthnumkanjinumeric>", "<$imperialera>",
   "<$imperialyear>", "<$imperialyear01>", "<$imperialyearkanjikazu>", "<$imperialyearspecialkanjikazu>", "<$imperialyearkanjinumeric>", "<$imperialyearspecialkanjinumeric>",
   "<$curpagenum>", "<$lastpagenum>", "<$volnum>", "<$chapnum>", "<$sectionnum>", "<$subsectionnum>", "<$chaptertitlename>", "<$FMuserdir>",
   "<$tblsheetnum>", "<$tblsheetcount>", "<$filename>", "<$fullfilename>"];

KLD_V.aBBgroups =        // --- building blocks in this group -------------------------------------
[
  ["user",    ["xxx"]]  // must be the first - filled in SwitchUIlanguage
, ["docinfo", ["<$volnum>", "<$chapnum>", "<$sectionnum>", "<$subsectionnum>", "<$chaptertitlename>", "<$FMuserdir>"]]
, ["file",    ["<$filename>", "<$fullfilename>"]]
, ["marker",  ["<$marker1>", "<$marker2>", "<$marker3>", "<$marker4>", "<$marker5>", "<$marker6>", "<$marker7>", "<$marker8>"]]
, ["pagenum", ["<$curpagenum>", "<$lastpagenum>"]]
, ["paranum", ["<$paranum[paratag]>", "<$paranumonly[paratag]>"]]
, ["paratag", ["<$paratag[paratag]>", "<$condtag[hitag, ..., lotag, nomatch]>"]]
, ["paratxt", ["<$paratext[paratag]>"]]
, ["tblsheet",["<$tblsheetnum>", "<$tblsheetcount>"]]
, ["time",    ["<$second>", "<$second00>", "<$minute>", "<$minute00>", "<$hour>", "<$hour01>", "<$hour24>", "<$ampm>", "<$AMPM>",
               "<$daynum>", "<$daynum01>", "<$dayname>", "<$shortdayname>", "<$monthnum>", "<$monthnum01>", "<$monthname>", "<$shortmonthname>", "<$year>", "<$shortyear>",
               "<$currenttime>", "<$modificationtime>", "<$creationtime>",
               "<$daynumkanjikazu>", "<$daynumkanjinumeric>", "<$monthnumkanjikazu>", "<$monthnumkanjinumeric>",
               "<$imperialera>", "<$imperialyear>", "<$imperialyear01>", "<$imperialyearkanjikazu>", "<$imperialyearspecialkanjikazu>", "<$imperialyearkanjinumeric>", "<$imperialyearspecialkanjinumeric>", ]]
];

KLD_V.aAllowedBB =      // ======== groups of allowed building blocks (order as in FM) ------------
[ ["Chapter Number",            ["user", "pagenum", "paranum", "docinfo", "time"]]
, ["Chapter Title Name",        ["user", "pagenum", "paranum", "docinfo", "time"]]
, ["Creation Date (Long)",      ["user", "time"]]
, ["Creation Date (Short)",     ["user", "time"]]
, ["Current Date (Long)",       ["user", "time"]]
, ["Current Date (Short)",      ["user", "time"]]
, ["Current Page #",            ["user", "pagenum", "paranum", "docinfo", "time"]]
, ["FM User Directory",         ["user", "pagenum", "paranum", "docinfo", "time"]]
, ["Filename (Long)",           ["user", "file",    "time"]]
, ["Filename (Short)",          ["user", "file",    "time"]]
, ["Modification Date (Long)",  ["user", "time"]]
, ["Modification Date (Short)", ["user", "time"]]
, ["Page Count",                ["user", "pagenum", "paranum", "docinfo", "time"]]
, ["Running H/F 1",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 10",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 11",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 12",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 13",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 14",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 15",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 16",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 17",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 18",            ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 2",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 3",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 4",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 5",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 6",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 7",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 8",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Running H/F 9",             ["user", "marker",  "paratxt", "paranum", "paratag", "time"]]
, ["Section Number",            ["user", "paranum", "time",    "pagenum", "docinfo"]]
, ["Sub Section Number",        ["user", "paranum", "time",    "pagenum", "docinfo"]]
, ["Table Continuation",        ["user", "time"]]
, ["Table Sheet",               ["user", "tblsheet","time"]]
, ["Volume Number",             ["user", "pagenum", "paranum", "docinfo", "time"]]
];

// === Settings for debugging =====================================================================
KLD_V.settings = {};
KLD_V.settings.bIsDebug = false;                      // standard debug output
KLD_V.settings.bVerbose = false;                      // detailed output

// === 'prototype' definitions ====================================================================
// Called by CollectVarsInDoc, GetVarNames
KLD_V.oVariable = function (sName, sValue, oVar, bSystem, oTr, sPageType, sPageName) {
  this.Name     =  sName;                         // Name of variable
  this.Value    =  sValue;                        // Contents of variable
  this.VarObj   =  oVar;                          // Var object in document
  this.IsSystem =  bSystem;                       // true for system variable
  this.TextRange=  oTr;                           // textrange of this occurance in the doc
} //--- end oVariable -----------------------------------------------------------------------------

